@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2015 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2022 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -35,25 +35,37 @@
  *   forward this exception.
  */
 
-#include "common.h"
 #include "prototypes.h"
 
+#ifdef USE_OS_THREADS
+THREAD_ID cron_thread_id=(THREAD_ID)0;
+#endif
+
 #ifdef USE_PTHREAD
 NOEXPORT void *cron_thread(void *arg);
 #endif
+
 #ifdef USE_WIN32
-NOEXPORT void cron_thread(void *arg);
+NOEXPORT unsigned __stdcall cron_thread(void *arg);
 #endif
-#if defined(USE_PTHREAD) || defined(USE_WIN32)
+
+#ifdef USE_OS_THREADS
 NOEXPORT void cron_worker(void);
+#ifndef OPENSSL_NO_DH
+#if OPENSSL_VERSION_NUMBER>=0x0090800fL
+NOEXPORT void cron_dh_param(BN_GENCB *);
+NOEXPORT BN_GENCB *cron_bn_gencb(void);
+NOEXPORT int bn_callback(int, int, BN_GENCB *);
+#else /* OpenSSL older than 0.9.8 */
 NOEXPORT void cron_dh_param(void);
-#endif
+NOEXPORT void dh_callback(int, int, void *);
+#endif /* OpenSSL 0.9.8 or later */
+#endif /* OPENSSL_NO_DH */
+#endif /* USE_OS_THREADS */
 
 #if defined(USE_PTHREAD)
 
 int cron_init() {
-    pthread_t thread;
-    pthread_attr_t pth_attr;
 #if defined(HAVE_PTHREAD_SIGMASK) && !defined(__APPLE__)
     sigset_t new_set, old_set;
 #endif /* HAVE_PTHREAD_SIGMASK && !__APPLE__*/
@@ -62,11 +74,10 @@ int cron_init() {
     sigfillset(&new_set);
     pthread_sigmask(SIG_SETMASK, &new_set, &old_set); /* block signals */
 #endif /* HAVE_PTHREAD_SIGMASK && !__APPLE__*/
-    pthread_attr_init(&pth_attr);
-    pthread_attr_setdetachstate(&pth_attr, PTHREAD_CREATE_DETACHED);
-    if(pthread_create(&thread, &pth_attr, cron_thread, NULL))
+    CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_THREAD_LIST]);
+    if(pthread_create(&cron_thread_id, NULL, cron_thread, NULL))
         ioerror("pthread_create");
-    pthread_attr_destroy(&pth_attr);
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_THREAD_LIST]);
 #if defined(HAVE_PTHREAD_SIGMASK) && !defined(__APPLE__)
     pthread_sigmask(SIG_SETMASK, &old_set, NULL); /* unblock signals */
 #endif /* HAVE_PTHREAD_SIGMASK && !__APPLE__*/
@@ -92,21 +103,28 @@ NOEXPORT void *cron_thread(void *arg) {
 #elif defined(USE_WIN32)
 
 int cron_init() {
-    if((long)_beginthread(cron_thread, 0, NULL)==-1)
-        ioerror("_beginthread");
+    CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_THREAD_LIST]);
+    cron_thread_id=(HANDLE)_beginthreadex(NULL, 0, cron_thread, NULL, 0, NULL);
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_THREAD_LIST]);
+    if(!cron_thread_id) {
+        ioerror("_beginthreadex");
+        return 1;
+    }
     return 0;
 }
 
-NOEXPORT void cron_thread(void *arg) {
+NOEXPORT unsigned __stdcall cron_thread(void *arg) {
     (void)arg; /* squash the unused parameter warning */
+
     tls_alloc(NULL, NULL, "cron");
     if(!SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_LOWEST))
         ioerror("SetThreadPriority");
     cron_worker();
-    _endthread(); /* it should never be executed */
+    _endthreadex(0); /* it should never be executed */
+    return 0;
 }
 
-#else /* !defined(USE_PTHREAD) && !defined(USE_WIN32) */
+#else /* USE_OS_THREADS */
 
 int cron_init() {
     /* not implemented for now */
@@ -118,19 +136,28 @@ int cron_init() {
 /* run the cron job every 24 hours */
 #define CRON_PERIOD (24*60*60)
 
-#if defined(USE_PTHREAD) || defined(USE_WIN32)
+#ifdef USE_OS_THREADS
 
 NOEXPORT void cron_worker(void) {
     time_t now, then;
     int delay;
+#if !defined(OPENSSL_NO_DH) && OPENSSL_VERSION_NUMBER>=0x0090800fL
+    BN_GENCB *bn_gencb;
+#endif
 
     s_log(LOG_DEBUG, "Cron thread initialized");
-    sleep(60); /* allow the other services to start with idle CPU */
+#if !defined(OPENSSL_NO_DH) && OPENSSL_VERSION_NUMBER>=0x0090800fL
+    bn_gencb=cron_bn_gencb();
+#endif
     time(&then);
     for(;;) {
         s_log(LOG_INFO, "Executing cron jobs");
 #ifndef OPENSSL_NO_DH
+#if OPENSSL_VERSION_NUMBER>=0x0090800fL
+        cron_dh_param(bn_gencb);
+#else /* OpenSSL older than 0.9.8 */
         cron_dh_param();
+#endif /* OpenSSL 0.9.8 or later */
 #endif /* OPENSSL_NO_DH */
         time(&now);
         s_log(LOG_INFO, "Cron jobs completed in %d seconds", (int)(now-then));
@@ -143,8 +170,8 @@ NOEXPORT void cron_worker(void) {
             then=now+delay;
         }
         s_log(LOG_DEBUG, "Waiting %d seconds", delay);
-        do { /* retry sleep() if it was interrupted by a signal */
-            sleep((unsigned)delay);
+        do { /* retry s_poll_sleep() if it was interrupted by a signal */
+            s_poll_sleep(delay, 0);
             time(&now);
             delay=(int)(then-now);
         } while(delay>0);
@@ -154,11 +181,16 @@ NOEXPORT void cron_worker(void) {
 }
 
 #ifndef OPENSSL_NO_DH
+
+#if OPENSSL_VERSION_NUMBER>=0x0090800fL
+NOEXPORT void cron_dh_param(BN_GENCB *bn_gencb) {
+#else /* OpenSSL older than 0.9.8 */
 NOEXPORT void cron_dh_param(void) {
+#endif /* OpenSSL 0.9.8 or later */
     SERVICE_OPTIONS *opt;
     DH *dh;
 
-    if(!dh_needed)
+    if(!dh_temp_params || !service_options.next)
         return;
 
     s_log(LOG_NOTICE, "Updating DH parameters");
@@ -169,33 +201,76 @@ NOEXPORT void cron_dh_param(void) {
         sslerror("DH_new");
         return;
     }
-    if(!DH_generate_parameters_ex(dh, 2048, 2, NULL)) {
+    if(!DH_generate_parameters_ex(dh, 2048, 2, bn_gencb)) {
         DH_free(dh);
         sslerror("DH_generate_parameters_ex");
         return;
     }
 #else /* OpenSSL older than 0.9.8 */
-    dh=DH_generate_parameters(2048, 2, NULL, NULL);
+    dh=DH_generate_parameters(2048, 2, dh_callback, NULL);
     if(!dh) {
         sslerror("DH_generate_parameters");
         return;
     }
-#endif
+#endif /* OpenSSL 0.9.8 or later */
 
     /* update global dh_params for future configuration reloads */
-    enter_critical_section(CRIT_DH); /* it only needs an rwlock here */
+    CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_DH]);
     DH_free(dh_params);
     dh_params=dh;
-    leave_critical_section(CRIT_DH);
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_DH]);
 
     /* set for all sections that require it */
+    CRYPTO_THREAD_read_lock(stunnel_locks[LOCK_SECTIONS]);
     for(opt=service_options.next; opt; opt=opt->next)
-        if(opt->option.dh_needed)
+        if(opt->option.dh_temp_params)
             SSL_CTX_set_tmp_dh(opt->ctx, dh);
+    CRYPTO_THREAD_unlock(stunnel_locks[LOCK_SECTIONS]);
     s_log(LOG_NOTICE, "DH parameters updated");
 }
+
+#if OPENSSL_VERSION_NUMBER>=0x0090800fL
+
+NOEXPORT BN_GENCB *cron_bn_gencb(void) {
+#if OPENSSL_VERSION_NUMBER>=0x10100000L
+    BN_GENCB *bn_gencb;
+
+    bn_gencb=BN_GENCB_new();
+    if(!bn_gencb) {
+        sslerror("BN_GENCB_new");
+        return NULL;
+    }
+    BN_GENCB_set(bn_gencb, bn_callback, NULL);
+    return bn_gencb;
+#else
+    static BN_GENCB bn_gencb;
+
+    BN_GENCB_set(&bn_gencb, bn_callback, NULL);
+    return &bn_gencb;
+#endif
+}
+
+NOEXPORT int bn_callback(int p, int n, BN_GENCB *cb) {
+    (void)p; /* squash the unused parameter warning */
+    (void)n; /* squash the unused parameter warning */
+    (void)cb; /* squash the unused parameter warning */
+    s_poll_sleep(0, 100); /* 100ms */
+    return 1; /* return nonzero for success */
+}
+
+#else /* OpenSSL older than 0.9.8 */
+
+NOEXPORT void dh_callback(int p, int n, void *arg) {
+    (void)p; /* squash the unused parameter warning */
+    (void)n; /* squash the unused parameter warning */
+    (void)arg; /* squash the unused parameter warning */
+    s_poll_sleep(0, 100); /* 100ms */
+}
+
+#endif /* OpenSSL 0.9.8 or later */
+
 #endif /* OPENSSL_NO_DH */
 
-#endif /* USE_PTHREAD || USE_WIN32 */
+#endif /* USE_OS_THREADS */
 
 /* end of cron.c */
